#!/usr/bin/env python


import logging
import os
import random
import re
import time
import urllib2

from sqlmap.common import Common
from sqlmap.google import Google
from sqlmap.parser import queriesParser
from sqlmap.settings import CSV_DIR
from sqlmap.settings import SUPPORTED_DBMS
from sqlmap.settings import VERSION


class Option(Common):
    """
    This class defines methods to check, parse and set options
    based upon command line parameters values and extra file

    @author: Bernardo Damele
    """

    def __init__(self, smDict):
        self.logger = logging.getLogger("sqlmapLog")
        self.args = smDict


    def __urllib2Opener(self):
        """
        This method creates and installs the urllib2 OpenerDirector
        globally
        """

        opener = urllib2.build_opener(self.proxyHandler, self.authHandler)
        self.args.opener = opener


    def __setGoogleDorking(self):
        """
        This method checks and set the Google dorking functionality
        """

        if not self.args.googleDork:
            return

        logMsg = "first request to Google to get the session cookie"
        self.logger.info(logMsg)

        googleObj = Google(self.proxyHandler)
        googleCookie = googleObj.getCookie()

        if googleCookie:
            googleCookie = googleCookie.groups()[0]

            logMsg = "sqlmap got '%s' as session cookie" % googleCookie
            self.logger.info(logMsg)
        else:
            errMsg  = "unable to parse Google response header "
            errMsg += "'set-cookie' to get the session cookie"
            raise Exception, errMsg

        matches = googleObj.search(self.args.googleDork)

        if not matches:
            errMsg  = "unable to find results for your "
            errMsg += "Google dork expression"
            raise Exception, errMsg

        self.args.testableHosts = googleObj.getTestableHosts()

        if self.args.testableHosts:
            logMsg  = "sqlmap got %d results for your " % len(matches)
            logMsg += "Google dork expression, "

            if len(matches) == len(self.args.testableHosts):
                logMsg += "all "
            else:
                logMsg += "%d " % len(self.args.testableHosts)

            logMsg += "of them are testable hosts"
            self.logger.info(logMsg)
        else:
            errMsg  = "sqlmap got %d results " % len(matches)
            errMsg += "for your Google dork expression, but none "
            errMsg += "of them has parameters to test for SQL "
            errMsg += "injection"
            raise Exception, errMsg


    def __setOutputResume(self):
        """
        This method checks and set the output text file and '--resume'
        command line option
        """

        if self.args.resume and not self.args.outputFile:
            warnMsg  = "you did not provide the text file to "
            warnMsg += "resume queries output from"
            self.logger.warn(warnMsg)
        elif self.args.outputFile:
            if os.path.exists(self.args.outputFile):
                self.args.writeFile = file(self.args.outputFile, "r")
                lines = self.args.writeFile.readlines()

                for line in lines:
                    if line.count("][") == 4:
                        line = line.split("][")

                        if len(line) != 5:
                            continue

                        url, place, parameter, expression, value = line

                        if not value:
                            continue

                        if url[0] == "[":
                            url = url[1:]

                        if value[-1] == "\n":
                            value = value[:-1]

                        if url not in self.args.resumedQueries.keys():
                            self.args.resumedQueries[url] = {}
                            self.args.resumedQueries[url][expression] = value
                        else:
                            if expression not in self.args.resumedQueries[url].keys():
                                self.args.resumedQueries[url][expression] = value
                            elif len(value) >= len(self.args.resumedQueries[url][expression]):
                                self.args.resumedQueries[url][expression] = value

                self.args.writeFile.close()

            try:
                self.args.writeFile = file(self.args.outputFile, "a")
                self.args.writeFile.write("\n[%s]\n" % time.strftime("%X %x"))
                self.args.writeFile.flush()
            except Exception, e:
                raise Exception, str(e)


    def __setRemoteDBMS(self):
        """
        This method checks the command line argument '--remote-dbms'
        value
        """

        if not self.args.dbms:
            return

        self.args.dbms = self.args.dbms.lower()
        self.args.MySQLVer = re.search("mysql ([\d\.]+)", self.args.dbms)

        if self.args.MySQLVer:
            self.args.dbms = "mysql"
            self.args.MySQLVer = self.args.MySQLVer.groups()[0]
        else:
            self.args.MySQLVer = None

        if self.args.dbms not in SUPPORTED_DBMS:
            errMsg  = "you provided an unsupported remote database "
            errMsg += "management system. The supported DBMS are "
            errMsg += "'MySQL', 'PostgreSQL' and 'Microsoft SQL "
            errMsg += "Server'. If you do not know the remote "
            errMsg += "DBMS, do not provide it and sqlmap will "
            errMsg += "fingerprint it for you"
            raise Exception, errMsg


    def __setHTTPProxy(self):
        """
        This method defines the HTTP proxy to pass by all HTTP requests
        """

        if not self.args.proxy:
            self.proxyHandler = urllib2.BaseHandler()
            return

        if not re.search("http://[\w\:\/\.\-\_]+\:[\d]+", self.args.proxy, re.I):
            errMsg  = "proxy value must be in format "
            errMsg += "'http://url:port'"
            raise Exception, errMsg
        else:
            proxyRegExp = re.search("([\w\:\/\.\-\_\:]+)\:([\d]+)", self.args.proxy, re.I)
            proxyUrl = proxyRegExp.groups()[0]
            proxyPort = proxyRegExp.groups()[1]
            self.proxyHandler = urllib2.ProxyHandler({"http": "%s:%s" % (proxyUrl, proxyPort)})


    def __setHTTPCookies(self):
        """
        This method set the HTTP Cookie header
        """

        if self.args.cookie:
            self.args.httpHeaders.append(("Connection", "Keep-Alive"))
            self.args.httpHeaders.append(("Cookie", self.args.cookie))


    def __setHTTPPasswordMgr(self, authRegExp):
        """
        This method initializes and returns an urllib2 HTTP
        authentication password manager
        """

        authUsername = authRegExp.groups()[0]
        authPassword = authRegExp.groups()[1]

        passwordMgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
        passwordMgr.add_password(None, self.args.link, authUsername, authPassword)

        return passwordMgr


    def __setHTTPAuthentication(self):
        """
        This method checks and set the HTTP authentication method
        (Basic or Digest), username and password
        """

        if not ( self.args.bAuth or self.args.dAuth ):
            self.authHandler = urllib2.BaseHandler()
            return

        if self.args.bAuth:
            bAuthRegExp = re.search("^(.*?)\:(.*?)$", self.args.bAuth)
        elif self.args.dAuth:
            dAuthRegExp = re.search("^(.*?)\:(.*?)$", self.args.dAuth)

        if ( self.args.bAuth and not bAuthRegExp ) or ( self.args.dAuth and not dAuthRegExp ):
            errMsg  = "HTTP Authentication value must be "
            errMsg += "in format 'username:password'"
            raise Exception, errMsg
        elif self.args.bAuth and bAuthRegExp:
            passwordMgr = self.__setHTTPPasswordMgr(bAuthRegExp)
            self.authHandler = urllib2.HTTPBasicAuthHandler(passwordMgr)
        elif self.args.dAuth and dAuthRegExp:
            passwordMgr = self.__setHTTPPasswordMgr(dAuthRegExp)
            self.authHandler = urllib2.HTTPDigestAuthHandler(passwordMgr)


    def __setCSVDir(self):
        if self.args.googleDork or ( not self.args.dumpTable and not self.args.dumpAll ):
            return

        self.args.csvPath = "%s%s%s" % (CSV_DIR, os.sep, self.args.link)

        if not os.path.isdir(CSV_DIR):
            os.mkdir(CSV_DIR, 0755)

        if not os.path.isdir(self.args.csvPath):
            os.mkdir(self.args.csvPath, 0755)


    def __setRequest(self):
        """
        This method checks, set the parameters and performs checks
        on 'data' command line parameter value for HTTP method POST
        """

        testableParams = False

        if self.args.googleDork:
            return

        self.forgeTestParameters()

        # Perform checks on GET parameters
        if "?" in self.args.url:
            self.args.url, parameters = self.splitUrlParams()
            self.args.parameters["GET"] = parameters

            paramDict = self.paramDict("GET", parameters)

            if paramDict:
                self.args.paramDict["GET"] = paramDict
                testableParams = True

        # Perform checks on POST parameters
        if self.args.data:
            self.args.parameters["POST"] = self.args.data

            paramDict = self.paramDict("POST", self.args.data)

            if paramDict:
                self.args.paramDict["POST"] = paramDict
                testableParams = True

        if self.args.httpMethod == "POST" and not self.args.data:
            errMsg = "HTTP POST method depends on '--data' value"
            raise Exception, errMsg

        # Perform checks on Cookie parameters
        if self.args.cookie:
            self.args.parameters["Cookie"] = self.args.cookie

            paramDict = self.paramDict("Cookie", self.args.cookie)

            if paramDict:
                self.args.paramDict["Cookie"] = paramDict
                testableParams = True

        # Perform checks on User-Agent parameters
        if self.args.uAgent:
            self.args.parameters["User-Agent"] = self.args.uAgent
            self.args.paramDict["User-Agent"] = { "": self.args.uAgent }
            testableParams = True

        if not testableParams:
            errMsg  = "all testable parameters you provided are not "
            errMsg += "present within the GET, POST and Cookie parameters"
            raise Exception, errMsg


    def __setHTTPMethod(self):
        """
        This method checks and set the HTTP method to perform HTTP
        requests with
        """

        if self.args.httpMethod:
            self.args.httpMethod = self.args.httpMethod.upper()

            if self.args.httpMethod not in ("GET", "POST"):
                warnMsg  = "'%s' " % self.args.httpMethod
                warnMsg += "is an unsupported HTTP method, "
                warnMsg += "setting to default method, GET"
                self.logger.warn(warnMsg)

                self.args.httpMethod = "GET"
        else:
            self.args.httpMethod = "GET"


    def __defaultHTTPUserAgent(self):
        """
        This method returns the default sqlmap HTTP User-Agent
        header string
        """

        return "sqlmap/%s (http://sqlmap.sourceforge.net)" % VERSION


    def __setHTTPUserAgent(self):
        """
        This method extracts a random HTTP User-Agent header string
        from a text file
        """

        if self.args.uAgent:
            self.args.httpHeaders.append(("User-Agent", self.args.uAgent))
            return

        if not self.args.userAgentsFile:
            self.args.httpHeaders.append(("User-Agent", self.__defaultHTTPUserAgent()))
            return

        logMsg  = "fetching random HTTP User-Agent header from "
        logMsg += "file '%s'" % self.args.userAgentsFile
        self.logger.info(logMsg)

        try:
            fd = open(self.args.userAgentsFile)
        except:
            warnMsg  = "unable to read HTTP User-Agent header "
            warnMsg += "file '%s'" % self.args.userAgentsFile
            self.logger.warn(warnMsg)

            self.args.httpHeaders.append(("User-Agent", self.__defaultHTTPUserAgent()))

            return

        count = 0
        userAgents = []

        while True:
            line = fd.readline()
            if not line:
                break
            userAgents.append(line)
            count += 1

        fd.close()

        if count == 1:
            userAgent = userAgents[0]
            userAgent = userAgent.replace("\n", "").replace("\r", "")
            self.args.httpHeaders.append(("User-Agent", userAgent))
        else:
            userAgent = userAgents[random.randint(0, count)]
            userAgent = userAgent.replace("\n", "").replace("\r", "")
            self.args.httpHeaders.append(("User-Agent", userAgent))

        logMsg  = "fetched random HTTP User-Agent header from "
        logMsg += "file '%s'" % self.args.userAgentsFile
        self.logger.info(logMsg)


    def __setURL(self):
        """
        This method checks and set the target url and port
        """

        if not self.args.url:
            return

        if not re.search("^http[s]*://[\w\.\-\_]+", self.args.url, re.I):
            port = re.search("^[\w\.\-\_]+:([\d]+)", self.args.url)

            if port:
                if int(port.groups()[0]) == 443:
                    self.args.url = "https://" + self.args.url
                else:
                    self.args.url = "http://" + self.args.url
            else:
                self.args.url = "http://" + self.args.url

        transport = re.search("^(http[s]*)://[\w\.\-\_]+", self.args.url, re.I)
        port = re.search("^http[s]*://[\w\.\-\_]+:([\d]+)", self.args.url, re.I)
        pageUrl = re.search("^http[s]*://[\w\.\-\_]+[:]*[\d]*(.*)", self.args.url, re.I)
        pageUrl = pageUrl.groups()[0]
        pageUrl = pageUrl.split("?")

        self.args.pageUrl = pageUrl[0]
        self.args.rootUrl = self.args.url[:self.args.url.index(self.args.pageUrl)]
        self.args.link = re.search("^http[s]*://([\w\.\-\_]+)", self.args.url, re.I).groups()[0]

        if transport:
            transport = transport.groups()[0]
        elif port:
            if int(port.groups()[0]) == 443:
                transport = "https"
            else:
                transport = "http"
        else:
            transport = "http"

        if port:
            self.args.port = int(port.groups()[0])
        elif transport == "http":
            self.args.port = 80
        else:
            self.args.port = 443


    def __cleanupParameters(self):
        """
        This method cleanup the input value of database, table and
        column values
        """

        if self.args.db:
            self.args.db = self.args.db.replace(", ", ",")

        if self.args.tbl:
            self.args.tbl = self.args.tbl.replace(", ", ",")

        if self.args.col:
            self.args.col = self.args.col.replace(", ", ",")


    def __setVariables(self):
        """
        This method set some needed variables
        """

        self.args.brackets = 0
        self.args.fingerprint = None
        self.args.htmlFp = []
        self.args.parameters = {}
        self.args.paramDict = {}
        self.args.injParameter = None
        self.args.injPlace = None
        self.args.absFilePaths = set()
        self.args.docRoot = None
        self.args.resumedQueries = {}
        self.args.unionComment = None
        self.args.unionCount = None
        self.args.unionPosition = None
        self.args.writeFile = None
        self.args.queries = queriesParser()
        self.args.httpHeaders = []


    def __setVerbose(self):
        """
        This method set the verbosity of sqlmap output messages
        """

        if not self.args.verbose:
            return

        self.args.verbose = int(self.args.verbose)

        if self.args.verbose == 1:
            self.logger.setLevel(logging.INFO)
        elif self.args.verbose > 1:
            if self.args.eta:
                self.args.verbose = 1
                self.logger.setLevel(logging.INFO)
            else:
                self.args.verbose = 2
                self.logger.setLevel(logging.DEBUG)


    def run(self):
        """
        This method checks, parses and set options based upon command
        line parameters values and extra file
        """

        self.__setVerbose()
        self.__setVariables()
        self.__cleanupParameters()
        self.__setURL()
        self.__setHTTPUserAgent()
        self.__setHTTPMethod()
        self.__setRequest()
        self.__setCSVDir()
        self.__setHTTPAuthentication()
        self.__setHTTPCookies()
        self.__setHTTPProxy()
        self.__setRemoteDBMS()
        self.__setOutputResume()
        self.__setGoogleDorking()
        self.__urllib2Opener()

        return self.args
