#!/usr/bin/env python


import re

from os.path import exists
from xml.sax import parse
from xml.sax.handler import ContentHandler

from sqlmap.settings import QUERIES_XML
from sqlmap.smdict import sqlmapDict


class queriesHandler(ContentHandler):
    """
    This class defines methods to parse the default DBMS queries
    from an XML file

    @author: Bernardo Damele
    """

    def __init__(self):
        self.queries = {}
        self.__dbms = ''
        self.__queriesDict = sqlmapDict()


    def startElement(self, name, attrs):
        if name == "dbms":
            data = sanitizeString(attrs.get("value"))
            self.__dbms = data

        elif name == "inference":
            data = sanitizeString(attrs.get("query"))
            self.__queriesDict.inference = data

        elif name == "banner":
            data = sanitizeString(attrs.get("query"))
            self.__queriesDict.banner = data

        elif name == "current_user":
            data = sanitizeString(attrs.get("query"))
            self.__queriesDict.currentUser = data

        elif name == "current_db":
            data = sanitizeString(attrs.get("query"))
            self.__queriesDict.currentDb = data

        elif name == "inband":
            self.__inband    = sanitizeString(attrs.get("query"))
            self.__inband2   = sanitizeString(attrs.get("query2"))
            self.__condition = sanitizeString(attrs.get("condition"))

        elif name == "blind":
            self.__blind  = sanitizeString(attrs.get("query"))
            self.__blind2 = sanitizeString(attrs.get("query2"))
            self.__count  = sanitizeString(attrs.get("count"))
            self.__count2 = sanitizeString(attrs.get("count2"))


    def endElement(self, name):
        if name == "dbms":
            self.queries[self.__dbms] = self.__queriesDict
            self.__queriesDict = sqlmapDict()

        elif name == "users":
            self.__users = {}
            self.__users["inband"] = { "query": self.__inband }
            self.__users["blind"]  = { "query": self.__blind, "count": self.__count }

            self.__queriesDict.users = self.__users

        elif name == "passwords":
            self.__passwords = {}
            self.__passwords["inband"] = { "query": self.__inband }
            self.__passwords["blind"]  = { "query": self.__blind, "count": self.__count }

            self.__queriesDict.passwords = self.__passwords

        elif name == "dbs":
            self.__dbs = {}
            self.__dbs["inband"] = { "query": self.__inband, "query2": self.__inband2 }
            self.__dbs["blind"]  = { "query": self.__blind, "query2": self.__blind2,
                                     "count": self.__count, "count2": self.__count2 }

            self.__queriesDict.dbs = self.__dbs

        elif name == "tables":
            self.__tables = {}
            self.__tables["inband"] = { "query": self.__inband, "condition": self.__condition }
            self.__tables["blind"]  = { "query": self.__blind, "count": self.__count }

            self.__queriesDict.tables = self.__tables

        elif name == "columns":
            self.__columns = {}
            self.__columns["inband"] = { "query": self.__inband }
            self.__columns["blind"]  = { "query": self.__blind, "query2": self.__blind2, "count": self.__count }

            self.__queriesDict.columns = self.__columns

        elif name == "dump_table":
            self.__dumpTable = {}
            self.__dumpTable["inband"] = { "query": self.__inband }
            self.__dumpTable["blind"]  = { "query": self.__blind, "count": self.__count }

            self.__queriesDict.dumpTable = self.__dumpTable


class htmlHandler(ContentHandler):
    """
    This class defines methods to parse the input HTML page to
    fingerprint the remote database management system

    @author: Bernardo Damele
    """

    def __init__(self, page):
        self.__dbms = None
        self.__page = page
        self.__regexp = None
        self.__match = None

        self.dbms = None


    def startElement(self, name, attrs):
        if name == "dbms":
            self.__dbms = attrs.get("value")

        if name == "error":
            self.__regexp = attrs.get("regexp")
            self.__match = re.search(self.__regexp, self.__page, re.I)

            if self.__match:
                self.dbms = self.__dbms
                self.__match = None


class bannerHandler(ContentHandler):
    """
    This class defines methods to parse and extract information from
    the given DBMS banner based upon the data in XML file

    @author: Bernardo Damele
    """

    def __init__(self, banner):
        self.__banner = banner
        self.__release = None
        self.__regexp = None
        self.__match = None

        self.release = None
        self.version = None
        self.servicepack = None


    def startElement(self, name, attrs):
        if name == "dbms":
            self.__release = attrs.get("release")
            self.__regexp = attrs.get("regexp")
            self.__match = re.search(self.__regexp, self.__banner)

        if name == "info" and self.__match:
            self.release = self.__release
            self.version = attrs.get("version")
            self.servicepack = attrs.get("servicepack")


def sanitizeString(string):
    """
    This function sanitizes the input string explicitly casting to a
    valid string and encoding in UTF-8
    """

    cleanString = str(string).encode('utf-8')

    return cleanString


def __sanitizeAndCheckXMLFile(xmlfile, string=None):
    """
    This function sanitizes the input string and checks the
    existance of XML file
    """

    if not exists(xmlfile):
        raise Exception, "unable to read XML file '%s'" % xmlfile

    if string:
        string = string.replace("\n", "").replace("\r", "")

        return string


def queriesParser():
    """
    This function calls a class to parse the default DBMS queries
    from an XML file
    """

    xmlfile = QUERIES_XML

    __sanitizeAndCheckXMLFile(xmlfile)
    handler = queriesHandler()
    parse(xmlfile, handler)

    return handler.queries


def htmlParser(page, xmlfile):
    """
    This function calls a class that parses the input HTML page to
    fingerprint the remote database management system
    """

    page = __sanitizeAndCheckXMLFile(xmlfile, page)
    handler = htmlHandler(page)
    parse(xmlfile, handler)

    return handler.dbms


def bannerParser(banner, xmlfile):
    """
    This function calls a class to extract information from the given
    DBMS banner based upon the data in XML file
    """

    banner = __sanitizeAndCheckXMLFile(xmlfile, banner)
    handler = bannerHandler(banner)
    parse(xmlfile, handler)

    return handler.release, handler.version, handler.servicepack

